==> prism-1.4.0-20250213.053831-7.jar: prism/internal/aws_sign.clj <==
(ns prism.internal.aws-sign
  "Modified from https://github.com/jacobemcken/aws-simple-sign"
  (:require
    [clojure.string :as str])
  (:import
    (java.net URI URL)
    (java.security MessageDigest)
    (java.time ZoneId ZoneOffset)
    (java.time.format DateTimeFormatter)
    (java.util Date)
    (javax.crypto Mac)
    (javax.crypto.spec SecretKeySpec)))

(defn ^:no-doc hash-sha256
  [^String input]
  (let [hash (MessageDigest/getInstance "SHA-256")]
    (.update hash (.getBytes input))
    (.digest hash)))

(def ^:no-doc digits
  (char-array "0123456789abcdef"))

(defn ^:no-doc hex-encode
  [bytes]
  (->> bytes
       (mapcat #(list (get digits (bit-shift-right (bit-and 0xF0 %) 4))
                      (get digits (bit-and 0x0F %))))))

(defn ^:no-doc hex-encode-str
  [bytes]
  (->> (hex-encode bytes)
       (apply str)))

;; Clojure implementation of signature
;; https://gist.github.com/souenzzo/21f3e81b899ba3f04d5f8858b4ecc2e9
;; https://github.com/joseferben/clj-aws-sign/ (ring middelware)

(defn ^:no-doc hmac-sha-256
  [key ^String data]
  (let [algo "HmacSHA256"
        mac (Mac/getInstance algo)]
    (.init mac (SecretKeySpec. key algo))
    (.doFinal mac (.getBytes data "UTF-8"))))

(defn ^:no-doc char-range
  [start end]
  (mapv char (range (int start) (inc (int end)))))

(def ^:no-doc unreserved-chars
  (->> (concat '(\- \. \_ \~)
               (char-range \A \Z)
               (char-range \a \z)
               (char-range \0 \9))
       (into #{})))

(def ^:no-doc url-unreserved-chars
  (conj unreserved-chars \/))

(defn ^:no-doc encode
  [skip-chars c]
  (if (skip-chars c)
    c
    (let [byte-val (int c)]
      (format "%%%X" byte-val))))

(defn ^:no-doc uri-encode
  [skip-chars uri]
  (->> (mapv (partial encode skip-chars) uri)
       (apply str)))

(def ^DateTimeFormatter ^:no-doc formatter
  (-> (DateTimeFormatter/ofPattern "yyyyMMdd'T'HHmmss'Z'")
      (.withZone (ZoneId/from ZoneOffset/UTC))))

(defn ^:no-doc compute-signature
  [{:keys [credentials str-to-sign region service short-date]}]
  (-> (str "AWS4" (:aws/secret-access-key credentials))
      (.getBytes)
      (hmac-sha-256 short-date)
      (hmac-sha-256 region)
      (hmac-sha-256 service)
      (hmac-sha-256 "aws4_request")
      (hmac-sha-256 str-to-sign)
      hex-encode-str))

(def ^:no-doc algorithm "AWS4-HMAC-SHA256")

(defn ^:no-doc ->query-str
  [query-params]
  (->> (into (sorted-map)
             (map (fn [[k v]] [(uri-encode unreserved-chars k) (uri-encode unreserved-chars v)]))
             query-params) ; sort AFTER URL encoding
       (mapv (fn [[k v]] (str k "=" v)))
       (str/join "&")))

(defn signature
  "AWS specification: https://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-header-based-auth.html

   Inspired by https://gist.github.com/souenzzo/21f3e81b899ba3f04d5f8858b4ecc2e9"
  [credentials canonical-url {:keys [scope method timestamp region service query-params content-sha256 signed-headers]}]
  (let [encoded-url (uri-encode url-unreserved-chars canonical-url)
        signed-headers (->> signed-headers
                            (map (fn [[k v]] [(str/lower-case k) v]))
                            (into (sorted-map)))
        headers-str (->> signed-headers
                         (map (fn [[k v]] (str k ":" (str/trim (or v "")) "\n")))
                         (apply str))
        signed-headers-str (str/join ";" (map key signed-headers))
        canonical-request (str (or method "GET") "\n"
                               encoded-url "\n"
                               (->query-str query-params) "\n"
                               headers-str "\n"
                               signed-headers-str "\n"
                               (or content-sha256 "UNSIGNED-PAYLOAD"))
        str-to-sign (str algorithm "\n"
                         timestamp "\n"
                         scope "\n"
                         (hex-encode-str (hash-sha256 canonical-request)))]
    (compute-signature {:credentials credentials
                        :str-to-sign str-to-sign
                        :region      region
                        :service     service
                        :short-date  (subs timestamp 0 8)})))

(defn ^:no-doc hashed-payload
  [payload]
  (when (or (nil? payload) (string? payload))
    (hex-encode-str (hash-sha256 (or payload "")))))

(defn sign-request
  "Takes a client and a Ring style request map.
   Returns an enriched Ring style map with the required headers
   needed for AWS signing."
  ([{:keys [body headers method url] :as request}
    {:keys [credentials ref-time region service query-params]
     :or   {service  "execute-api"
            ref-time (Date.)}}]
   (let [query-params (into (sorted-map) query-params)
         timestamp (.format formatter (.toInstant ^Date ref-time))
         url-obj (-> (URI. url)
                     .toURL)
         content-sha256 (hashed-payload body)
         session-token (:aws/session-token credentials)
         signed-headers (assoc headers
                          "Host" (.getHost url-obj)
                          "x-amz-content-sha256" content-sha256
                          "x-amz-date" timestamp
                          "x-amz-security-token" session-token)
         scope (str (subs timestamp 0 8) "/" region "/" service "/aws4_request")
         signature-str (signature credentials
                                  (.getPath url-obj)
                                  {:scope          scope
                                   :timestamp      timestamp
                                   :region         region
                                   :service        service
                                   :method         (-> method name str/upper-case)
                                   :signed-headers signed-headers
                                   :query-params   (update-keys query-params name)
                                   :content-sha256 content-sha256})]
     (-> request
         (assoc :headers (cond-> (dissoc signed-headers "Host")
                                 (nil? session-token) (dissoc "x-amz-security-token")) ; overwrite headers to include necessary x-amz-* ones
                :query-params query-params)
         (update :headers assoc
                 "Authorization" (str algorithm " Credential=" (:aws/access-key-id credentials) "/" scope ", "
                                      "SignedHeaders=" (str/join ";" (map key signed-headers)) ", "
                                      "Signature=" signature-str))))))

==> prism-1.4.0-20250214.013829-9.jar: prism/internal/aws_sign.clj <==
(ns prism.internal.aws-sign
  "Modified from https://github.com/jacobemcken/aws-simple-sign"
  (:require
    [clojure.string :as str])
  (:import
    (java.net URI)
    (java.security MessageDigest)
    (java.time ZoneId ZoneOffset)
    (java.time.format DateTimeFormatter)
    (java.util Date)
    (javax.crypto Mac)
    (javax.crypto.spec SecretKeySpec)))

(defn ^:no-doc hash-sha256
  [^String input]
  (let [hash (MessageDigest/getInstance "SHA-256")]
    (.update hash (.getBytes input))
    (.digest hash)))

(def ^:no-doc digits
  (char-array "0123456789abcdef"))

(defn ^:no-doc hex-encode
  [bytes]
  (->> bytes
       (mapcat #(list (get digits (bit-shift-right (bit-and 0xF0 %) 4))
                      (get digits (bit-and 0x0F %))))))

(defn ^:no-doc hex-encode-str
  [bytes]
  (->> (hex-encode bytes)
       (apply str)))

;; Clojure implementation of signature
;; https://gist.github.com/souenzzo/21f3e81b899ba3f04d5f8858b4ecc2e9
;; https://github.com/joseferben/clj-aws-sign/ (ring middelware)

(defn ^:no-doc hmac-sha-256
  [key ^String data]
  (let [algo "HmacSHA256"
        mac (Mac/getInstance algo)]
    (.init mac (SecretKeySpec. key algo))
    (.doFinal mac (.getBytes data "UTF-8"))))

(defn ^:no-doc char-range
  [start end]
  (mapv char (range (int start) (inc (int end)))))

(def ^:no-doc unreserved-chars
  (->> (concat '(\- \. \_ \~)
               (char-range \A \Z)
               (char-range \a \z)
               (char-range \0 \9))
       (into #{})))

(def ^:no-doc url-unreserved-chars
  (conj unreserved-chars \/))

(defn ^:no-doc encode
  [skip-chars c]
  (if (skip-chars c)
    c
    (let [byte-val (int c)]
      (format "%%%X" byte-val))))

(defn ^:no-doc uri-encode
  [skip-chars uri]
  (->> (mapv (partial encode skip-chars) uri)
       (apply str)))

(def ^DateTimeFormatter ^:no-doc formatter
  (-> (DateTimeFormatter/ofPattern "yyyyMMdd'T'HHmmss'Z'")
      (.withZone (ZoneId/from ZoneOffset/UTC))))

(defn ^:no-doc compute-signature
  [{:keys [credentials str-to-sign region service short-date]}]
  (-> (str "AWS4" (:aws/secret-access-key credentials))
      (.getBytes)
      (hmac-sha-256 short-date)
      (hmac-sha-256 region)
      (hmac-sha-256 service)
      (hmac-sha-256 "aws4_request")
      (hmac-sha-256 str-to-sign)
      hex-encode-str))

(def ^:no-doc algorithm "AWS4-HMAC-SHA256")

(defn signature
  "AWS specification: https://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-header-based-auth.html

   Inspired by https://gist.github.com/souenzzo/21f3e81b899ba3f04d5f8858b4ecc2e9"
  [credentials canonical-url {:keys [scope method timestamp region service query-string content-sha256 signed-headers]}]
  (let [encoded-url (uri-encode url-unreserved-chars canonical-url)
        signed-headers (into (sorted-map) (map (fn [[k v]] [(str/lower-case k) v])) signed-headers)
        headers-str (->> signed-headers
                         (map (fn [[k v]] (str k ":" (str/trim (or v "")) "\n")))
                         (apply str))
        signed-headers-str (str/join ";" (map key signed-headers))
        canonical-request (str (or method "GET") "\n"
                               encoded-url "\n"
                               query-string "\n"
                               headers-str "\n"
                               signed-headers-str "\n"
                               (or content-sha256 "UNSIGNED-PAYLOAD"))
        str-to-sign (str algorithm "\n"
                         timestamp "\n"
                         scope "\n"
                         (hex-encode-str (hash-sha256 canonical-request)))]
    (compute-signature {:credentials credentials
                        :str-to-sign str-to-sign
                        :region      region
                        :service     service
                        :short-date  (subs timestamp 0 8)})))

(defn ^:no-doc hashed-payload
  [payload]
  (when (or (nil? payload) (string? payload))
    (hex-encode-str (hash-sha256 (or payload "")))))

(defn sign-request
  "Takes a client and a Ring style request map.
   Returns an enriched Ring style map with the required headers
   needed for AWS signing."
  ([{:keys [body headers request-method url query-string] :as request} ;assumes query-string is built from sorted map
    {:keys [credentials ref-time region service]
     :or   {service  "execute-api"
            ref-time (Date.)}}]
   (let [timestamp (.format formatter (.toInstant ^Date ref-time))
         url-obj (-> (URI. url)
                     .toURL)
         port (.getPort url-obj)
         content-sha256 (hashed-payload body)
         session-token (:aws/session-token credentials)
         signed-headers (cond-> (assoc headers
                                  "Host" (cond-> (.getHost url-obj)
                                                 (pos? port) (str \: port))
                                  "x-amz-content-sha256" content-sha256
                                  "x-amz-date" timestamp)
                                session-token (assoc "x-amz-security-token" session-token))
         scope (str (subs timestamp 0 8) "/" region "/" service "/aws4_request")
         signature-str (signature credentials
                                  (.getPath url-obj)
                                  {:scope          scope
                                   :timestamp      timestamp
                                   :region         region
                                   :service        service
                                   :method         (-> request-method name str/upper-case)
                                   :signed-headers signed-headers
                                   :query-string   query-string
                                   :content-sha256 content-sha256})
         headers (-> (dissoc signed-headers "Host")
                     (assoc "Authorization" (str algorithm " Credential=" (:aws/access-key-id credentials) "/" scope ", "
                                                 "SignedHeaders=" (str/join ";" (map key signed-headers)) ", "
                                                 "Signature=" signature-str))
                     (update-keys str/lower-case))]
     (assoc request :headers headers))))
